using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;
using System.Windows.Controls;
using log4net;
using System;
using Loki.Bot;
using Loki.Game;
using Loki.Game.GameData;
using Loki.Game.Objects;
using Utility = Loki.Bot.Logic.Bots.OldGrindBot.Utility;
using Loki.Bot.Logic.Bots.OldGrindBot;
using Loki.Common;
using Buddy.Coroutines;


namespace CorruptedAreaTweaks
{
    internal class CorruptedAreaTweaks : IPlugin
    {
        private static readonly ILog Log = Logger.GetLoggerInstanceForType();
        public static TaskManager GrindBotTaskManager = GetGrindBotTaskManager();
        private TaskManager _taskManager;       
       

        public static event Action<uint, uint, DatWorldAreaWrapper> AreaChanged;
        public static event Action PlayerDied;

        private bool _enabled;

        /// <summary>User's can change this to make the plugin mover closer or further away from a cached location.</summary>
        private const int Range = 15;

        /// <summary> The name of the plugin. </summary>
        public string Name
        {
            get { return "CorruptedAreaTweaks"; }
        }
        public void Deinitialize()
        {
            Log.Debug("[coraptareatweks] Deinitialize");
        }

        /// <summary> The description of the plugin. </summary>
        public string Description
        {
            get { return "A plugin that provides some additional functionality in Corrupted areas."; }
        }

        /// <summary>The author of the plugin.</summary>
        public string Author
        {
            get { return "Bossland GmbH"; }
        }

        /// <summary>The version of the plugin.</summary>
        public string Version
        {
            get { return "vers 1.0"; }
        }

        /// <summary>Initializes this plugin.</summary>
        public void Initialize()
        {
            Log.DebugFormat("[CorruptedAreaTweaks] Initialize");
        }

        /// <summary> The plugin start callback. Do any initialization here. </summary>
        public void Start()
        {                      
            AddTask(new TrackUniques(), "ExploreTask", front: false, after: false, before: true);
            Log.DebugFormat("[CorruptedAreaTweaks] added before ExploreTask");

            _taskManager.Remove("ExplorationCompleteTask (Early)");
        
        }
    
        public async Task<bool> Logic(string type, params dynamic[] param)
        {
           
            return false;
        }

        public static TaskManager GetGrindBotTaskManager()
        {
            var bot = BotManager.CurrentBot;
            return (TaskManager)bot.Execute("GetTaskManager");
        }
        public static void AddTask(ITask task, string taskToProcess = "", bool front = true, bool after = false, bool before = false)
        {
            if (front)
            {
                if (!GrindBotTaskManager.AddAtFront(task))
                    Log.ErrorFormat("[{0}] AddAtFront failed for {1}", "AddTask", task.Name);
                else
                    Log.DebugFormat("[{0}] {1} Sucessfully added at front of the list", "AddTask", task.Name);
            }

            if (after)
            {
                if (!GrindBotTaskManager.AddAfter(task, taskToProcess))
                    Log.ErrorFormat("[{0}] AddAfter failed for {1}", "AddTask", task.Name);
                else
                    Log.DebugFormat("[{0}] {1} Sucessfully added after {2}", "AddTask", task.Name, taskToProcess);
            }

            if (before)
            {
                if (!GrindBotTaskManager.AddBefore(task, taskToProcess))
                    Log.ErrorFormat("[{0}] AddBefore failed for {1}", "AddTask", task.Name);
                else
                    Log.DebugFormat("[{0}] {1} Sucessfully added before {2}", "AddTask", task.Name, taskToProcess);
            }
        }
        /// <summary> The plugin tick callback. Do any update logic here. </summary>
        public void Tick()
        {
        }

        /// <summary> The plugin stop callback. Do any pre-dispose cleanup here. </summary>
        public void Stop()
        {
            Log.DebugFormat("[CorruptedAreaTweaks] Stop");
        }

        #region Implementation of IConfigurable

        public JsonSettings Settings
        {
            get { return null; }
        }

        /// <summary> The plugin's settings control. This will be added to the Exilebuddy Settings tab.</summary>
        public UserControl Control
        {
            get { return null; }
        }

        #endregion

        #region Implementation of IEnableable

        /// <summary>Is this plugin currently enabled?</summary>
        public bool IsEnabled
        {
            get { return _enabled; }
        }

        /// <summary> The plugin is being enabled.</summary>
        public void Enable()
        {
            Log.DebugFormat("[CorruptedAreaTweaks] Enable");
            _enabled = true;
        }

        /// <summary> The plugin is being disabled.</summary>
        public void Disable()
        {
            Log.DebugFormat("[CorruptedAreaTweaks] Disable");
            _enabled = false;
        }
        public object Execute(string name, params dynamic[] param)
        {
            return null;
        }

        #endregion

        #region Implementation of IDisposable

        /// <summary> </summary>
        public void Dispose()
        {
        }

        #endregion

        #region Override of Object

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return Name + ": " + Description;
        }

        #endregion

        public class TrackUniques : ITask
        {
            private readonly Stopwatch _sw = new Stopwatch();
            private readonly Dictionary<int, Vector2i> _uniques = new Dictionary<int, Vector2i>();

            private bool _enabled = true;

            /// <summary>The name of this task.</summary>
            public string Name
            {
                get { return "TrackUniques"; }
            }

            /// <summary>A description of what this task does.</summary>
            public string Description
            {
                get { return "This task helps BasicGrindBot seek out unique bosses in Corrupted areas."; }
            }

            /// <summary>The author of this task.</summary>
            public string Author
            {
                get { return "Bossland GmbH"; }
            }

            /// <summary>The version of this task.</summary>
            public string Version
            {
                get  { return "Unfinished"; }
            }

            /// <summary>
            /// The tasks's execution logic.
            /// </summary>
            /// <returns>true if the TaskManager should not execute any other tasks, and false if it should.</returns>
            public async Task<bool> Logic(string type, params dynamic[] param)
            {

                if (type == "core_area_changed_event")
                {
                    uint oldSeed = (uint)param[0];
                    uint newSeed = (uint)param[1];
                    var newArea = (DatWorldAreaWrapper)param[3];                    
                    EventInvocators.RaiseAreaChangedEvent(oldSeed, newSeed, newArea);
                    _uniques.Clear();
                }
                if (type == "core_player_died_event")
                {
                    EventInvocators.RaisePlayerDiedEvent();
                }
                if (type != "task_execute") return false;
                
                if (!LokiPoe.CurrentWorldArea.IsCorruptedArea)
                    return false;

                await Coroutine.Sleep(Utility.LatencySafeValue(50));
                // No uniques to process.
                if (_uniques.Count == 0)
                    return false;

                // Get the an unique.
                var kvp = _uniques.First();

                // If we aren't in range of it, move towards it so we can trigger it.
                if (LokiPoe.Me.Position.Distance(kvp.Value) > Range)
                {
                    if (!PlayerMover.MoveTowards(kvp.Value))
                    {
                        Log.ErrorFormat("[TrackUniques] MoveTowards failed for {0}.", kvp.Value);
                    }
                    return true;
                }

                // We are now close enough to the position, check to see if the object exists or is dead.
                var m = LokiPoe.ObjectManager.GetObjectById<Monster>(kvp.Key);
                if (m == null || m.IsDead)
                {
                    _uniques.Remove(kvp.Key);
                }
                else
                {
                    _uniques[kvp.Key] = m.Position;
                }

                // We now are relying on the CR to kill the boss. If it doesn't, we'll just die or get
                // stuck around where it spawns.

                return true;
            }

            /// <summary>The bot Start event.</summary>
            public void Start()
            {
                
            }

            /// <summary>The bot Tick event.</summary>
            public void Tick()
            {
                if (!LokiPoe.IsInGame || LokiPoe.Me.IsInTown || LokiPoe.Me.IsInHideout)
                    return;

                if (!_sw.IsRunning)
                {
                    _sw.Start();
                }
                else
                {
                    if (_sw.ElapsedMilliseconds < 250)
                        return;
                }

                var uniques =
                    LokiPoe.ObjectManager.Objects.OfType<Monster>()
                        .Where(
                            m =>
                                !AreaStateCache.Current.IsBlacklisted(m.Id) &&
                                ( /*m.IsStrongboxMinion ||*/ m.Rarity == Rarity.Unique) && !m.IsDead &&
                                m.Reaction == Reaction.Enemy);
                foreach (var unique in uniques)
                {
                    if (!_uniques.ContainsKey(unique.Id))
                    {
                        Log.DebugFormat("[TrackUniques] Now adding {0} at {1}.", unique.Name, unique.Position);
                        _uniques.Add(unique.Id, unique.Position);
                    }
                }

                _sw.Restart();
            }

            public static class EventInvocators
            {
                                
                public static void RaiseAreaChangedEvent(uint oldSeed, uint newSeed, DatWorldAreaWrapper newArea)
                {
                    var handler = AreaChanged;
                    if (handler != null) handler(oldSeed, newSeed, newArea);
                   // _uniques.Clear();
                }

                public static void RaisePlayerDiedEvent()
                {
                    var handler = PlayerDied;
                    if (handler != null) handler();
                }
            }
            /// <summary>The bot Stop event.</summary>
            public void Stop()
            {
                
            }

            /// <summary>Is this task currently enabled?</summary>
            public bool IsEnabled
            {
                get { return _enabled; }
            }

            /// <summary>Called when the task should be enabled.</summary>
            public void Enable()
            {
                _enabled = true;
            }

            /// <summary>Called when the task should be disabled.</summary>
            public void Disable()
            {
                _enabled = false;
            }
            public object Execute(string name, params dynamic[] param)
            {
                return null;
            }
            
        }
    }
}
